import datetime
import pdb

import shopify
import pyactiveresource
import shopify_python_api_patcher
from shopify_python_api_patcher.utils import get_current_release

from em_tasks import logger
from em_tasks.utils.sku_parser import SkuParser
from em_tasks.utils.exchange_rate import ExchangeRate
from em_tasks.store.models import *


class OrderUtil():
  def __init__(self, store_name, app_token, release=None):
    self.store_name = store_name
    self.app_token = app_token
    if release:
      self.release = release
    else:
      self.release = get_current_release()

    shop_url = "{}.myshopify.com".format(store_name)
    self.shopify_credentials = {'domain': shop_url, 'version': self.release, 'token': self.app_token}

    session = shopify.Session(shop_url, self.release, self.app_token)
    shopify.ShopifyResource.activate_session(session)
    self.shop = shopify.Shop.current()

  def import_orders(self, params={}):
    for order in self.list_orders(params):
      order_attrs = self.format_order(order)
      if not order_attrs:
        continue

      # Order already canceled, skip
      if order.cancelled_at and order_attrs['cancelled_at']:
        continue

      # Order already fulfilled, skip
      if order.fulfillment_status == 'fulfilled' and order_attrs['fulfillment_status'] == 'fulfilled':
        continue

      try:
        order_db = Order.get(Order.s_order_id == order.id)
        for k, v in order_attrs.items():
          setattr(order_db, k, v)

        order_db.save()
      except Order.DoesNotExist:
        Order.insert(order_attrs).execute()
      except Exception as e:
        logger.exception(e)

      order_items = self.format_order_items(order)
      if not order_items:
        continue

      order_items_mapping = dict()
      for order_item in order_items:
        order_item_key = '{}-{}'.format(order_item['s_order_id'], order_item['s_line_item_id'])
        order_items_mapping[order_item_key] = order_item

      order_items_db = OrderItem.select().where(OrderItem.s_order_id == order.id).execute()
      for order_item_db in order_items_db:
        order_item_key = '{}-{}'.format(order_item_db.s_order_id, order_item_db.s_line_item_id)
        if order_item_key not in order_items_mapping:
          continue

        order_item = order_items_mapping.pop(order_item_key)
        for k, v in order_item.items():
          setattr(order_item_db, k, v)

        order_item_db.save()

      if order_items_mapping:
        OrderItem.insert_many(list(order_items_mapping.values())).execute()

  def cancel_risk_orders(self, params={}):
    for order in self.list_orders(params):
      if order.cancelled_at:
        logger.debug('[OrderAlreadyCanceled] %s', order.name)
        continue

      if order.financial_status in ['voided', 'refunded']:
        logger.debug('[OrderVoidedOrRefunded] %s', order.name)
        continue

      should_cancel_order = False
      order_risks = shopify.OrderRisk.find(order_id=order.id)
      for order_risk in order_risks:
        if order_risk.recommendation == 'cancel':
          should_cancel_order = True
          break

      if not should_cancel_order:
        continue

      order.cancel(reason='fraud', email=True)

  def capture_fulfillable_orders(self):
    orders = dict()
    for order in Order.select(Order).join(OrderItem, on=(Order.s_order_id == OrderItem.s_order_id)).where((Order.financial_status == 'authorized') & (OrderItem.fulfillable == 1)).execute():
      orders[order.s_order_id] = order

    order_ids = list(orders.keys())
    chunks = [order_ids[x:x + 50] for x in range(0, len(order_ids), 50)]
    for chunk in chunks:
      ids = ','.join([str(oid) for oid in chunk])
      for shopify_order in shopify.Order.find(ids=ids):
        if shopify_order.financial_status == 'authorized':
          succeed = shopify_order.capture()
          if succeed:
            logger.info('[OrderCaptured] Order: %s, OrderNumber: %s', shopify_order.id, shopify_order.name)
          else:
            logger.info('[OrderCaptureFailed] Order: %s, OrderNumber: %s', shopify_order.id, shopify_order.name)

  def capture_fulfilled_orders(self, days=7):
    today = datetime.datetime.utcnow()
    from_date = today - datetime.timedelta(days=days)

    s_order_ids = dict()
    for order_item in OrderItem.select(OrderItem).join(FulfillOrder, on=(OrderItem.id == FulfillOrder.order_item_id)).where(OrderItem.s_purchase_date > from_date).execute():
      s_order_ids[order_item.s_order_id] = None

    if not s_order_ids:
      return

    s_order_ids = list(s_order_ids.keys())
    orders = dict()
    for order in Order.select().where((Order.s_order_id.in_(s_order_ids)) & (Order.financial_status == 'authorized')).execute():
      orders[order.s_order_id] = order

    order_ids = list(orders.keys())
    chunks = [order_ids[x:x + 50] for x in range(0, len(order_ids), 50)]
    for chunk in chunks:
      ids = ','.join([str(oid) for oid in chunk])
      for shopify_order in shopify.Order.find(ids=ids):
        if shopify_order.financial_status == 'authorized':
          succeed = shopify_order.capture()
          if succeed:
            logger.info('[OrderCaptured] Order: %s, OrderNumber: %s', shopify_order.id, shopify_order.name)
          else:
            logger.info('[OrderCaptureFailed] Order: %s, OrderNumber: %s', shopify_order.id, shopify_order.name)

  def upload_trackings(self, days=14):
    for order_item in OrderItem.select(OrderItem, FulfillOrder).join(FulfillOrder, on=(OrderItem.id == FulfillOrder.order_item_id), attr='fulfill_order').where((FulfillOrder.stc == 1) & (FulfillOrder.buy_tracking != '') & (FulfillOrder.buy_tracking.is_null(False)) & (OrderItem.s_fulfillment_id.is_null(True))).execute():
      fulfillment_order_id = order_item.s_fulfillment_order_id
      fulfillment_line_item_id = None
      try:
        # Shopify python API bug
        fo_prefix = shopify.FulfillmentOrders._prefix_source
        shopify.FulfillmentOrders._prefix_source = ''

        fulfillment_order = shopify.FulfillmentOrders.find(str(fulfillment_order_id))
        for li in fulfillment_order.line_items:
          if li.line_item_id == order_item.s_line_item_id:
            fulfillment_line_item_id = li.id
            break

        shopify.FulfillmentOrders._prefix_source = fo_prefix
      except Exception as e:
        logger.exception(e)

      quantity = order_item.s_quantity
      carrier = order_item.fulfill_order.buy_carrier
      tracking = order_item.fulfill_order.buy_tracking
      fulfillment = shopify.Fulfillment()
      fulfillment.notify_customer = True
      fulfillment.tracking_info = {
        'company': carrier,
        'number': tracking
      }
      if carrier == 'PARCELSAPP':
        fulfillment.tracking_info['url'] = 'https://parcelsapp.com/en/tracking/{}'.format(tracking)

      fo = [{
        'fulfillment_order_id': fulfillment_order_id,
      }]
      if fulfillment_line_item_id:
        fo = [{
          'fulfillment_order_id': fulfillment_order_id,
          'fulfillment_order_line_items': [{
            'id': fulfillment_line_item_id,
            'quantity': quantity
          }]
        }]

      fulfillment.line_items_by_fulfillment_order = fo
      try:
        # Shopify python API bug
        f_prefix = shopify.Fulfillment._prefix_source
        shopify.Fulfillment._prefix_source = ''

        succeed = fulfillment.save()

        shopify.Fulfillment._prefix_source = f_prefix

        if succeed:
          order_item.s_fulfillment_id = fulfillment.id
          order_item.s_carrier = carrier
          order_item.s_tracking = tracking
          order_item.save()
          logger.info(
            '[OrderTrackingUploaded] OrderNumber: %s, FulfillmentID: %s, Carrier: %s, Tracking: %s',
            order_item.s_order_number, order_item.s_fulfillment_id, order_item.s_carrier, order_item.s_tracking)
      except Exception as e:
        pdb.set_trace()
        logger.exception(e)
        logger.debug(fulfillment.to_json())

  @classmethod
  def find_amz_order_offers(self, offer_service, days=7):
    sources = ['AMZ_{}'.format(marketplace) for marketplace in SkuParser.supported_marketplaces]
    today = datetime.datetime.utcnow()
    from_date = today - datetime.timedelta(days=days)
    order_items = OrderItem.select().where((OrderItem.risky == 0) & (OrderItem.canceled == 0) & (OrderItem.source.in_(sources)) & OrderItem.source_price.is_null() & (OrderItem.s_purchase_date > from_date)).execute()

    order_items_by_m = dict()
    for order_item in order_items:
      marketplace = order_item.source.split('_').pop().lower()
      if marketplace not in order_items_by_m:
        order_items_by_m[marketplace] = dict()

      order_items_by_m[marketplace][order_item.source_id] = order_item

    for marketplace, order_items in order_items_by_m.items():
      asins = list(order_items.keys())
      offers = offer_service.get_offers(marketplace.lower(), asins)
      if not offers:
        continue

      for asin, offer in offers.items():
        if asin not in order_items:
          continue

        if not offer:
          continue

        order_item = order_items[asin]

        source_price_in_usd = offer['price']
        currency = offer.get('currency', 'USD')
        if currency != 'USD':
          exchange_rate = ExchangeRate.get_exchange_rate('USD', currency)
          source_price_in_usd = offer['price'] / exchange_rate

        order_item.source_price = round(source_price_in_usd, 2)
        order_item.estimate_profit = round(order_item.s_price - source_price_in_usd, 2)
        if order_item.estimate_profit > 0:
          order_item.fulfillable = True
        order_item.save()
        logger.info(
          '[OrderOfferFind] Order: %s, Price: %s, Source: %s, SourceId: %s, SourcePrice: %s',
          order_item.s_order_number, order_item.s_price, order_item.source, order_item.source_id, order_item.source_price)

  def list_orders(self, params={}):
    days = 14
    today = datetime.datetime.utcnow()
    from_date = today - datetime.timedelta(days=days)
    from_date_s = from_date.replace(microsecond=0).isoformat()
    page_size = 250
    dft_params = {'limit': page_size, 'processed_at_min': from_date_s, 'status': 'any'}
    filters = dict(dft_params)
    filters.update(params)

    next_page_url = None
    has_more_orders = True
    with shopify.Session.temp(**self.shopify_credentials):
      while has_more_orders:
        if next_page_url:
          orders = shopify.Order.find(from_=next_page_url)
        else:
          orders = shopify.Order.find(**filters)

        if not orders:
          has_more_orders = False
          break

        print('[OrdersNextPage] {}'.format(orders.next_page_url))
        if len(orders) < page_size:
          has_more_orders = False
        else:
          if next_page_url == orders.next_page_url:
            break

          next_page_url = orders.next_page_url

        for order in orders:
          yield order

  def format_order(self, shopify_order):
    if shopify_order.test:
      return

    shipping_charge = round(sum([float(shipping_line.discounted_price) for shipping_line in shopify_order.shipping_lines]), 2)
    order = {
      's_order_id': shopify_order.id,
      's_order_number': shopify_order.name,
      'purchase_date': shopify_order.processed_at,
      'total': round(float(shopify_order.total_price), 2),
      'item_total': shopify_order.total_line_items_price,
      'tax': shopify_order.total_tax,
      'shipping_charge': shipping_charge,
      'discount': shopify_order.total_discounts,
      'cancelled_at': shopify_order.cancelled_at,
      'cancel_reason': shopify_order.cancel_reason,
      'closed_at': shopify_order.closed_at,
      'financial_status': shopify_order.financial_status,
      'fulfillment_status': shopify_order.fulfillment_status,
      'note': shopify_order.note,
      'order_status_url': shopify_order.order_status_url
    }
    if not shopify_order.refunds:
      return order

    refund_amount = 0
    for refund in shopify_order.refunds:
      refund_amount += sum([float(transaction.amount) for transaction in refund.transactions])
    refund = shopify_order.refunds.pop()
    order['refunded_at'] = refund.processed_at
    order['refund_amount'] = refund_amount

    return order

  def format_order_items(self, shopify_order):
    if shopify_order.test:
      return

    order_items = []

    sa = shopify_order.shipping_address
    shipping_address = {
      'first_name': sa.first_name,
      'last_name': sa.last_name,
      'address1': sa.address1,
      'address2': sa.address2,
      'company': sa.company,
      'city': sa.city,
      'state': sa.province_code,
      'zipcode': sa.zip,
      'country': sa.country_code,
      'phone': sa.phone,
    }
    ba = shopify.billing_address
    billing_address = {
      'first_name': ba.first_name,
      'last_name': ba.last_name,
      'address1': ba.address1,
      'address2': ba.address2,
      'company': ba.company,
      'city': ba.city,
      'state': ba.province_code,
      'zipcode': ba.zip,
      'country': ba.country_code,
      'phone': ba.phone,
    }

    shipping_line = shopify_order.shipping_lines[0]
    fulfillments = shopify_order.fulfillments
    fulfillment_orders = shopify.FulfillmentOrders.find(order_id=shopify_order.id)
    canceled = bool(shopify_order.cancelled_at)

    order_risks = shopify.OrderRisk.find(order_id=shopify_order.id)
    refunds = shopify_order.refunds

    for line_item in shopify_order.line_items:
      vendor_id = 1
      source = None
      source_id = None
      product_id = line_item.product_id
      store_product = StoreProduct.get_or_none(StoreProduct.s_product_id == product_id)
      if store_product:
        vendor_id = store_product.store_seller_id
        source = store_product.source
        source_id = store_product.source_product_id
      else:
        sku = line_item.sku
        parsed_vendor_id, source, source_id = SkuParser.parse(sku)
        if parsed_vendor_id:
          vendor_id = parsed_vendor_id

      product = None
      url = None
      image_url = None
      options_text = None
      if product_id:
        try:
          product = shopify.Product.find(product_id)
          url = 'https://{}/products/{}?variant_id={}'.format(self.shop.domain, product.handle, line_item.variant_id)
          for image in product.images:
            if line_item.variant_id in image.variant_ids:
              image_url = image.src
              break
          if not image_url and product.images:
            image_url = product.images[0].src

          variant = None
          options_text = ''
          for v in product.variants:
            if v.id == line_item.variant_id:
              variant = v
              break
          if variant:
            if variant.option1 != 'Default Title':
              option_texts = []
              for option in product.options:
                option_text = '{}:{}'.format(option.name, getattr(variant, 'option{}'.format(option.position)))
                option_texts.append(option_text)

              options_text = ';'.join(option_texts)
        except pyactiveresource.connection.ResourceNotFound:
          pass

      price = round(float(line_item.price), 2)
      tax = round(sum([float(tax_line.price) for tax_line in line_item.tax_lines]), 2)
      discount = round(float(line_item.total_discount), 2)
      total = price + tax

      order_item = {
        'email': shopify_order.email,
        'vendor_id': vendor_id,
        's_order_id': shopify_order.id,
        's_order_number': shopify_order.name,
        's_purchase_date': shopify_order.processed_at,
        's_total': total,
        's_price': price,
        's_tax': tax,
        's_discount': discount,
        's_quantity': line_item.quantity,
        's_line_item_id': line_item.id,
        's_product_id': line_item.product_id,
        's_variant_id': line_item.variant_id,
        's_name': line_item.name,
        's_sku': line_item.sku,
        's_url': url,
        's_image_url': image_url,
        's_options_text': options_text,
        's_shipping_method': shipping_line.code,
        's_shipping_charge': round(float(shipping_line.discounted_price), 2),
        'source': source,
        'source_id': source_id,
        'canceled': canceled,
        'note': shopify_order.note,
        'created_at': shopify_order.created_at,
        'updated_at': shopify_order.updated_at
      }

      order_item.update(shipping_address)
      order_item.billing_address = json.dumps(billing_address)

      fulfillment_order_selected = None
      for fulfillment_order in fulfillment_orders:
        for fli in fulfillment_order.line_items:
          if fli.line_item_id == line_item.id:
            fulfillment_order_selected = fulfillment_order
            break

        if fulfillment_order_selected:
          break

      if fulfillment_order_selected:
        order_item['s_fulfillment_order_id'] = fulfillment_order_selected.id
        delivery_method = fulfillment_order_selected.delivery_method
        if delivery_method.min_delivery_date_time:
          order_item['s_min_delivery_date'] = delivery_method.min_delivery_date_time

        if delivery_method.max_delivery_date_time:
          order_item['s_max_delivery_date'] = delivery_method.max_delivery_date_time

      fulfillment_selected = None
      for fulfillment in fulfillments:
        for fli in fulfillment.line_items:
          if fli.id == line_item.id or fli.fulfillment_line_item_id == line_item.id:
            fulfillment_selected = fulfillment
            break

        if fulfillment_selected:
          break

      if fulfillment_selected:
        order_item['s_fulfillment_id'] = fulfillment_selected.id
        order_item['s_carrier'] = fulfillment_selected.tracking_company
        order_item['s_tracking'] = fulfillment_selected.tracking_number
        # order_item['s_tracking_url'] = fulfillment_selected.tracking_url
        # order_item['s_shipping_cost']
        order_item['status'] = fulfillment_selected.status
      else:
        order_item['status'] = shopify_order.financial_status

      order_risk_selected = None
      for order_risk in order_risks:
        if order_risk.recommendation != 'accept':
          order_risk_selected = order_risk
          break

      if not order_risk_selected and order_risks:
        order_risk_selected = order_risks[0]

      if order_risk_selected:
        order_item['risky'] = order_risk_selected.recommendation != 'accept'
        order_item['risky_action'] = order_risk_selected.recommendation
        order_item['risky_message'] = order_risk_selected.message

      refund_selected = None
      refund_line_item_selected = None
      for refund in refunds:
        for refund_line_item in refund.refund_line_items:
          if refund_line_item.line_item_id == line_item.id:
            refund_line_item_selected = refund_line_item
            break

        if refund_line_item_selected:
          refund_selected = refund
          break

      if refund_line_item_selected:
        order_item['refund_qty'] = refund_line_item_selected.quantity
        order_item['refund_amount'] = round(float(refund_line_item_selected.subtotal) + float(refund_line_item_selected.total_tax), 2)
        order_item['refund_note'] = refund_selected.note

      order_items.append(order_item)

    return order_items
