import shopify
import pyactiveresource
import shopify_python_api_patcher
from shopify_python_api_patcher.utils import get_current_release

from em_tasks import logger


class ProductUtil():
  def __init__(self, store_name, app_token, release=None):
    self.store_name = store_name
    self.app_token = app_token
    if release:
      self.release = release
    else:
      self.release = get_current_release()

    shop_url = "{}.myshopify.com".format(store_name)
    self.shopify_credentials = {'domain': shop_url, 'version': self.release, 'token': self.app_token}

    session = shopify.Session(shop_url, self.release, self.app_token)
    shopify.ShopifyResource.activate_session(session)
    self.shop = shopify.Shop.current()

  def list_products(self, next_page_url=None):
    page_size = 250
    next_page_url = next_page_url
    has_more_products = True
    with shopify.Session.temp(**self.shopify_credentials):
      while has_more_products:
        if next_page_url:
          products = shopify.Product.find(from_=next_page_url)
        else:
          products = shopify.Product.find(limit=page_size)

        if not products:
          has_more_products = False
          break

        print('[ProductsNextPage] %s', products.next_page_url)
        if len(products) < page_size:
          has_more_products = False
        else:
          if next_page_url == products.next_page_url:
            break

          next_page_url = products.next_page_url

        for product in products:
          yield product

  def delete_product(self, product_id):
    with shopify.Session.temp(**self.shopify_credentials):
      product = shopify.Product.find(product_id)
      if not product:
        return

      product.destroy()

  def create_product(self, prod):
    sp = None
    with shopify.Session.temp(**self.shopify_credentials):
      p_attrs = self.to_shopify_product(prod)
      variants = p_attrs.pop('variants', [])
      if not variants:
        return

      sp = shopify.Product()
      for k, v in p_attrs.items():
        setattr(sp, k, v)

      variant_invs = dict()
      svs = []
      for variant in variants:
        v_quantity = variant.pop('inventory_quantity')
        v_location_id = variant.pop('location_id')
        variant_invs[variant['sku']] = {'location_id': v_location_id, 'quantity': v_quantity}

        sv = shopify.Variant()
        for k, v in variant.items():
          setattr(sv, k, v)
        svs.append(sv)
      sp.variants = svs
      product_created = sp.save()

      if not product_created:
        return

      for variant in sp.variants:
        if variant.sku not in variant_invs:
          continue

        v_inv = variant_invs[variant.sku]
        quantity = v_inv['quantity']
        location_id = v_inv['location_id']
        if location_id:
          inventory_level = shopify.InventoryLevel.connect(
            location_id, variant.inventory_item_id, relocate_if_necessary=True)
          inventory_level.set(
            inventory_level.location_id, inventory_level.inventory_item_id, quantity,
            disconnect_if_necessary=True)
        else:
          ils = shopify.InventoryLevel.find(inventory_item_ids=str(variant.inventory_item_id))
          if ils:
            for il in ils:
              shopify.InventoryLevel.set(il.location_id, il.inventory_item_id, quantity)

    return sp

  def set_products_offer(self, products):
    with shopify.Session.temp(**self.shopify_credentials):
      product_ids = ','.join([str(pid) for pid in list(products.keys())])
      sps = shopify.Product.find(ids=product_ids)
      for sp in sps:
        if sp.id not in products:
          continue

        product = products[sp.id]
        offer = product.get('offer', None)
        # Offer service encountered errors
        if isinstance(offer, bool) and not offer:
          continue

        # Product is not available on offer service
        if offer is None:
          continue

        # Product with multiple variants should supply each variant's offer
        variants = list(sp.variants)
        if len(variants) > 1:
          if not isinstance(offer, dict):
            continue
        else:
          v = variants[0]
          if v.id not in offer:
            offer = {v.id: offer}

        for v in variants:
          vid = v.id
          if vid not in offer:
            continue

          v_offer = offer[vid]
          if not v_offer:
            continue

          src_offer = v_offer.get('src_offer', None)
          price = v_offer['price']
          quantity = v_offer['quantity']
          price_v = round(float(v.price), 2)
          quantity_v = int(v.inventory_quantity)
          if price <= 0 or quantity <= 0:
            if quantity_v <= 0:
              # Already out of stock
              logger.debug(
                '[OfferUnchanged] ProductId: %s, VariantId: %s, VariantPrice: %s, VariantQuantity: %s, Price: %s, Quantity: %s',
                sp.id, v.id, price_v, quantity_v, price, quantity)
              continue

            ils = shopify.InventoryLevel.find(inventory_item_ids=str(v.inventory_item_id))
            if ils:
              for il in ils:
                if il.location_id == self.shop.primary_location_id:
                  # Don't manage primary location inventory, this location used for manage physical inventory
                  continue

                shopify.InventoryLevel.set(il.location_id, il.inventory_item_id, 0)
                logger.debug(
                  '[VariantInventoryChanged] ProductId: %s, VariantId: %s, LocationId: %s, InventoryItemId: %s, Quantity: %s',
                  sp.id, v.id, il.location_id, il.inventory_item_id, 0)
          else:
            if abs(price - price_v) > 1:
              # No need to change price when price changes in 1 USD
              v.price = price
              v.save()
              logger.debug(
                '[VariantPriceChanged] ProductId: %s, VariantId: %s, PriceNew: %s, PriceOld: %s, SrcOffer: %s',
                sp.id, v.id, price, price_v, src_offer)
            else:
              logger.debug(
                '[VariantPriceUnchanged] ProductId: %s, VariantId: %s, PriceNew: %s, PriceOld: %s, SrcOffer: %s',
                sp.id, v.id, price, price_v, src_offer)

            if (quantity > 0 and quantity_v < 10) or (quantity <= 0 and quantity_v > 0):
              # No need to change quantity when quantity is still more than 10
              ils = shopify.InventoryLevel.find(inventory_item_ids=str(v.inventory_item_id))
              if ils:
                for il in ils:
                  if il.location_id == self.shop.primary_location_id:
                    # Don't manage primary location inventory, this location used for manage physical inventory
                    continue

                  shopify.InventoryLevel.set(il.location_id, il.inventory_item_id, quantity)
                  logger.debug(
                    '[VariantInventoryChanged] ProductId: %s, VariantId: %s, LocationId: %s, InventoryItemId: %s, Quantity: %s',
                    sp.id, v.id, il.location_id, il.inventory_item_id, quantity)

  def update_products(self, products):
    with shopify.Session.temp(**self.shopify_credentials):
      product_ids = ','.join([str(pid) for pid in list(products.keys())])
      sps = shopify.Product.find(ids=product_ids)
      for sp in sps:
        if sp.id not in products:
          continue

        product = products[sp.id]
        data = product.get('data', None)
        if not data:
          continue

        title = data.get('title_en', None) or data['title']
        if title and sp.title != title:
          sp.title = title

        description = data.get('description', None)
        if description and sp.description != description:
          sp.body_html = description

        sp.save()

        prod_images = [image.strip() for image in data.get('images', '').split(';') if image.strip()]
        if prod_images:
          for image in sp.images:
            image.destroy()

          for img_url in prod_images:
            image = shopify.Image()
            image.product_id = sp.id
            image.src = img_url
            image.save()

  def to_shopify_product(self, prod):
    title = prod.get('title_en', None) or prod['title']
    p_attrs = {
      'published_scope': 'web',
      'status': 'active',
      'title': title,
      'body_html': prod['description'],
    }
    brand = prod.get('brand', None)
    if brand:
      p_attrs['vendor'] = brand

    slug = prod.get('slug', None)
    if slug:
      p_attrs['handle'] = slug

    categories = prod.get('categories', '')
    if categories:
      p_attrs['product_type'] = categories

    images = []
    prod_images = [image.strip() for image in prod.get('images', '').split(';')]
    for img_url in prod_images:
      if not img_url:
        continue

      images.append({'src': img_url})
    if not images:
      return

    p_attrs['images'] = images

    metafields = []
    for spec in prod.get('specifications', []):
      metafields.append({
        'key': spec['name'],
        'value': str(spec['value']),
        'type': 'single_line_text_field',
        'namespace': 'global'
      })
    if metafields:
      p_attrs['metafields'] = metafields

    options = prod.get('options', [])
    if len(options) > 3:
      return

    options_mapping = dict()
    for idx, option in enumerate(options):
      option_key = 'option{}'.format(idx + 1)
      p_attrs[option_key] = option['name']
      options_mapping[option['name']] = option_key

    prod_variants = prod.pop('variants', [])
    if prod_variants:
      variants = []
      for pv in prod_variants:
        v = {
          'sku': pv['sku'],
          'price': round(float(pv['price']), 2),
          'inventory_quantity': pv.get('quantity', 50),
          'location_id': pv.get('location_id', None),
          'fulfillment_service': 'manual',
          'inventory_management': 'shopify',
          'taxable': True,
          'requires_shipping': True,
        }

        option_values = pv.get('option_values', [])
        if not option_values:
          continue

        invalid_variant = False
        for ov in option_values:
          if ov['option_name'] not in options_mapping:
            invalid_variant = True
            break

          option_key = options_mapping[ov['option_name']]
          v[option_key] = ov['option_value']

        if invalid_variant:
          continue

        barcode = pv.get('barcode', None)
        if barcode:
          v['barcode'] = barcode

        images = []
        v_images = [image.strip() for image in pv.get('images', '').split(';')]
        for img_url in v_images:
          if not img_url:
            continue

          images.append({'src': img_url})

        if images:
          v['images'] = images

        variants.append(v)

      if variants:
        p_attrs['variants'] = variants
    else:
      v = {
        'sku': prod['sku'],
        'price': round(float(prod['price']), 2),
        'inventory_quantity': prod.get('quantity', 50),
        'location_id': prod.get('location_id', None),
        'fulfillment_service': 'manual',
        'inventory_management': 'shopify',
        'taxable': True,
        'requires_shipping': True
      }

      upc = prod.get('upc', None)
      if upc:
        v['barcode'] = upc

      weight = prod.get('weight', None)
      if weight:
        v['weight'] = weight
        v['weight_unit'] = 'lb'

      p_attrs['variants'] = [v]

    return p_attrs

  def to_gmc_product(self, product):
    pass
